pub mod message;

#[macro_use]
extern crate derivative;

use std::{
  io::Cursor,
  mem,
  net::{Ipv4Addr, Ipv6Addr, SocketAddr},
  sync::Arc,
};

use automerge::{
  sync::{State as SyncState, SyncDoc},
  Automerge,
};
use axum::{
  extract::{
    ws::{Message as WsMessage, WebSocket, WebSocketUpgrade},
    State,
  },
  response::Response,
  routing::get,
  Router, Server,
};
use chrono::Utc;
use clap::Parser;
use dashmap::{DashMap, DashSet};
use flume::r#async::SendSink;
use frontend::frontend;
use futures::{stream, FutureExt, SinkExt, StreamExt};
use message::Message;
use mzlib::axum_error::Result;
use tokio::join;
use uuid::Uuid;

#[derive(Debug, Parser)]
struct Opt {
  #[clap(long, default_value = "6106")]
  port: u16,
}

#[derive(Default, Clone)]
struct AppState {
  clients: Clients,
  rooms: Rooms,
}

#[derive(Default, Clone)]
struct Clients(Arc<DashMap<Uuid, Client>>);

struct Client {
  writer: SendSink<'static, Result<axum::extract::ws::Message, axum::Error>>,
  rooms: DashSet<Uuid>,
}

#[derive(Default, Clone)]
struct Rooms(Arc<DashMap<Uuid, Room>>);

#[derive(Default)]
struct Room {
  document: Automerge,
  sync_state: SyncState,
  connected_clients: DashSet<Uuid>,
  current_uploader: Option<Uuid>,
}

impl Room {
  pub fn split_borrow(&mut self) -> (&mut Automerge, &mut SyncState) {
    (&mut self.document, &mut self.sync_state)
  }
}

#[tokio::main]
async fn main() -> Result<()> {
  let opts = Opt::parse();

  let state = AppState::default();

  let app = Router::new()
    .route("/ws", get(handler))
    .fallback(frontend)
    .with_state(state);

  let service = app.into_make_service();

  let addr6 = SocketAddr::from((Ipv6Addr::UNSPECIFIED, opts.port));
  println!("Listening on {}...", addr6);

  let server6 = Server::bind(&addr6).serve(service);

  server6.await?;

  Ok(())
}

#[allow(non_upper_case_globals)]
mod frontend {
  use axum::{
    body::{Body, HttpBody},
    http::{header::CONTENT_TYPE, HeaderMap, HeaderValue, StatusCode, Uri},
    response::IntoResponse,
  };
  use rust_embed::RustEmbed;

  #[derive(RustEmbed)]
  #[folder = "dist/"]
  struct Frontend;

  pub async fn frontend(uri: Uri) -> impl IntoResponse {
    let path = uri.path().trim_start_matches("/").to_owned();

    match Frontend::get(&path).or_else(|| Frontend::get("index.html")) {
      Some(file) => {
        let guess = mime_guess::from_path(path);
        let mut headers = HeaderMap::default();
        if let Some(guess) = guess.first() {
          headers.append(
            CONTENT_TYPE,
            HeaderValue::from_str(&guess.to_string()).unwrap(),
          );
        }

        let data = file.data.as_ref();
        let body = Body::from(data.to_owned());
        let body = body.boxed();
        (headers, body).into_response()
      }
      None => {
        (StatusCode::NOT_FOUND, format!("No route for {}", uri)).into_response()
      }
    }
  }
}

async fn handler(
  ws: WebSocketUpgrade,
  State(state): State<AppState>,
) -> Response {
  ws.on_upgrade(|socket| handle_socket(state, socket).map(|res| res.unwrap()))
}

async fn handle_socket(state: AppState, socket: WebSocket) -> Result<()> {
  let (socket_tx, socket_rx) = socket.split();

  // Generate an ID for this connection
  let client_id = Uuid::new_v4();
  println!("Connected client {client_id:?}");

  // First, let's create multiplexed versions of these
  let mut socket_tx = {
    let (tx, rx) = flume::unbounded();
    tokio::spawn(rx.into_stream().forward(socket_tx));
    tx.into_sink()
  };
  let mut socket_rx = {
    let (tx, rx) = flume::unbounded();
    tokio::spawn(
      socket_rx
        .forward(tx.into_sink().sink_map_err(|err| axum::Error::new(err))),
    );
    rx.into_stream()
  };

  // Now register this client into the set of clients
  let rooms = DashSet::default();
  let client = Client {
    writer: socket_tx.clone(),
    rooms: rooms.clone(),
  };
  state.clients.0.insert(client_id, client);

  // Send a hello message
  socket_tx
    .send(Ok(
      Message::ServerHello {
        timestamp: Utc::now(),
        client_id: client_id.clone(),
      }
      .into_ws_message()?,
    ))
    .await?;

  // Wait for messages
  while let Some(msg) = socket_rx.next().await {
    match msg {
      WsMessage::Binary(ref bin) => {
        let cursor = Cursor::new(bin);
        let message = ciborium::from_reader::<Message, _>(cursor)?;
        // println!("received message: {message:?}");

        match &message {
          Message::Automerge {
            client_id,
            room_id,
            message: inner_message,
          } => {
            let mut room = state
              .rooms
              .0
              .entry(room_id.clone())
              .or_insert_with(|| Room::default());
            rooms.insert(room_id.clone());

            // Update local state
            {
              let (document, sync_state) = room.split_borrow();
              document
                .receive_sync_message(sync_state, inner_message.clone().0)?;
            }
            // println!("inner doc: {:?}", room.document);

            // Remove current client, so send to everyone else
            let connected_clients = room.connected_clients.clone();
            connected_clients.remove(client_id);

            // Send to everyone else
            let message = message.clone();
            broadcast(&state, connected_clients, message).await;
          }

          Message::JoinRoom { room_id } => {
            let room = state
              .rooms
              .0
              .entry(room_id.clone())
              .or_insert_with(|| Room::default());
            rooms.insert(room_id.clone());

            // Add to the list of connected clients
            room.connected_clients.insert(client_id);
            let connected_clients = room.connected_clients.clone();
            mem::drop(room);
            println!("Added to room");

            // Tell each clients which clients are connected
            broadcast(
              &state,
              connected_clients.clone(),
              Message::RoomClientList {
                room_id: room_id.clone(),
                clients: connected_clients.into_iter().collect(),
              },
            )
            .await;
          }

          Message::ChatMessage {
            timestamp: _,
            message_id: _,
            room_id,
            author: _,
            content: _,
          } => {
            let room = state
              .rooms
              .0
              .entry(room_id.clone())
              .or_insert_with(|| Room::default());

            // Add to the list of connected clients
            let connected_clients = room.connected_clients.clone();
            mem::drop(room);
            println!("Added to room");

            // Broadcast to the room
            broadcast(&state, connected_clients.clone(), message).await;
          }

          Message::FileUploadBegin { room_id, .. }
          | Message::FileUploadComplete { room_id }
          | Message::FileUploadChunk { room_id, .. } => {
            let mut room = state
              .rooms
              .0
              .entry(room_id.clone())
              .or_insert_with(|| Room::default());

            let connected_clients = room.connected_clients.clone();
            connected_clients.remove(&client_id);

            match &message {
              Message::FileUploadBegin { .. } => match room.current_uploader {
                Some(v) if v == client_id => {}
                Some(_v) => continue,
                None => room.current_uploader = Some(client_id.clone()),
              },
              Message::FileUploadComplete { .. } => {
                room.current_uploader = None;
              }
              Message::FileUploadChunk { .. } => {}
              _ => {}
            };

            broadcast(&state, connected_clients, message).await;
          }

          _ => {}
        };
      }
      _ => {}
    }
  }

  state.clients.0.remove(&client_id);
  println!("Disconnecting client {client_id:?}");

  // Tell other clients about the disconnection
  stream::iter(rooms)
    .for_each(|room_id| {
      let room = state
        .rooms
        .0
        .entry(room_id.clone())
        .or_insert_with(|| Room::default());

      room.connected_clients.remove(&client_id);
      let connected_clients = room.connected_clients.clone();
      mem::drop(room);

      let room_id = room_id.clone();
      {
        let state_ref = &state;
        async move {
          broadcast(
            state_ref,
            connected_clients.clone(),
            Message::RoomClientList {
              room_id: room_id,
              clients: connected_clients.into_iter().collect(),
            },
          )
          .await;
        }
      }
    })
    .await;

  Ok(())
}

async fn broadcast<I>(state: &AppState, client_ids: I, message: Message)
where
  I: IntoIterator<Item = Uuid>,
{
  let client_ids = client_ids.into_iter().collect::<Vec<_>>();
  // println!("Broadcasting message to {} clients:", client_ids.len());
  // println!("  - message: {message:?}");

  stream::iter(
    client_ids
      .into_iter()
      .filter_map(|client_id| state.clients.0.get(&client_id))
      .map(|client| client.writer.clone()),
  )
  .for_each_concurrent(None, move |mut writer| {
    let message = message.clone();
    async move {
      writer.send(Ok(message.into_ws_message().unwrap())).await;
    }
  })
  .await;
}
