use std::collections::HashSet;
use std::fmt::Formatter;
use std::io::Cursor;

use axum::extract::ws;
use chrono::{DateTime, Utc};

use mzlib::axum_error::Result;
use serde::de::Visitor;
use serde::{Deserialize, Deserializer, Serialize, Serializer};
use uuid::Uuid;

#[derive(Derivative, Clone, Serialize, Deserialize)]
#[derivative(Debug)]
#[serde(tag = "type")]
pub enum Message {
  ServerHello {
    timestamp: DateTime<Utc>,
    client_id: Uuid,
  },
  JoinRoom {
    room_id: Uuid,
  },
  RoomClientList {
    room_id: Uuid,
    clients: HashSet<Uuid>,
  },
  ChatMessage {
    timestamp: DateTime<Utc>,
    message_id: Uuid,
    room_id: Uuid,
    author: Uuid,
    content: String,
  },
  Automerge {
    client_id: Uuid,
    room_id: Uuid,
    #[derivative(Debug = "ignore")]
    message: AutomergeMessage,
  },
  FileUploadBegin {
    room_id: Uuid,
    #[serde(with = "serde_bytes")]
    header: Vec<u8>,
  },
  FileUploadChunk {
    room_id: Uuid,
    #[serde(with = "serde_bytes")]
    data: Vec<u8>,
  },
  FileUploadComplete {
    room_id: Uuid,
  },
}

impl Message {
  pub fn into_ws_message(self) -> Result<ws::Message> {
    let vec = Vec::new();
    let mut cursor = Cursor::new(vec);
    ciborium::into_writer(&self, &mut cursor)?;
    let vec = cursor.into_inner();
    Ok(ws::Message::Binary(vec))
  }
}

#[derive(Clone)]
pub struct AutomergeMessage(pub automerge::sync::Message);

impl Serialize for AutomergeMessage {
  fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
  where
    S: Serializer,
  {
    let bytes = self.0.clone().encode();
    serializer.serialize_bytes(&bytes)
  }
}

impl<'de> Deserialize<'de> for AutomergeMessage {
  fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
  where
    D: Deserializer<'de>,
  {
    struct ThisVisitor;
    impl<'v> Visitor<'v> for ThisVisitor {
      type Value = AutomergeMessage;

      fn expecting(&self, f: &mut Formatter) -> std::fmt::Result {
        write!(f, "an automerge sync message")
      }

      fn visit_bytes<E>(self, v: &[u8]) -> Result<Self::Value, E>
      where
        E: serde::de::Error,
      {
        automerge::sync::Message::decode(v)
          .map(AutomergeMessage)
          .map_err(|_err| E::custom("invalid automerge sync message"))
      }
    }
    let visitor = ThisVisitor;
    deserializer.deserialize_bytes(visitor)
  }
}
